# TODO: more descriptive errors
# FIXME: fetch_* is a shit name
from netbox_proxmox_sync.api.config import NB_API
from netbox_proxmox_sync.api.utils.errors import APIError, ValidationError
from netbox_proxmox_sync.api.proxmox.extract import Proxmox
from .delete import all as delete_everything


def _assert_nodes_exist(proxmox_nodes):
    node_names = set([node['name'] for node in proxmox_nodes])
    existing_names = set([
        node.name for node in NB_API.dcim.devices.filter(name=node_names)
    ])
    missing_nodes = list(node_names - existing_names)
    if len(missing_nodes) > 0:
        missing = [f'Node "{node_name}" is missing!' for node_name in missing_nodes]
        raise ValidationError('Not all cluster nodes registered in NetBox!', missing)


def _assert_cluster_does_not_exist(proxmox_cluster):
    cluster_name = proxmox_cluster['name']
    netbox_cluster_list = NB_API.virtualization.clusters.filter(name=cluster_name)
    if len(netbox_cluster_list) != 0:
        raise ValidationError(f'Virtualization cluster "{cluster_name}" already exists!')


def cluster(proxmox_cluster, proxmox_nodes):
    try:
        # Already formatted the data for this :)
        netbox_cluster = NB_API.virtualization.clusters.create(**proxmox_cluster)
    except Exception as e:
        raise APIError(e)
    # Update nodes' cluster
    node_names = [node['name'] for node in proxmox_nodes]
    netbox_nodes = list(NB_API.dcim.devices.filter(name=[node_names]))
    for node in netbox_nodes:
        node.cluster = {'name': proxmox_cluster['name']}
    if not NB_API.dcim.devices.update(netbox_nodes):
        # In case of error "rollback"
        netbox_cluster.delete()
        raise APIError('Failed to set Nodes\' cluster!')
    # Return JSON serializable dict :)
    return dict(netbox_cluster)


def tags():
    proxmox_tags = Proxmox.fetch_tags()
    try:
        existing_tags = NB_API.extras.tags.filter(content_types=[
            'virtualization.virtualmachines',
        ])
    except Exception as e:
        raise APIError(e)
    # Tags for VMs currently should only be managed by this plugin
    # TODO: we could add a custom field to mark tags managed by the plugin, for instance
    if len(existing_tags) > 0:
        names = [tag.name for tag in existing_tags]
        errors = [f'Tag "{name}" should not exist!' for name in names]
        raise ValidationError('Some VM tags already exist!', errors)
    try:
        netbox_tags = NB_API.extras.tags.create(proxmox_tags)
    except Exception as e:
        raise APIError(e)
    # Return JSON serializable list :)
    return [dict(tag) for tag in netbox_tags]


def custom_fields():
    try:
        vmid_already_exists = len(NB_API.extras.custom_fields.filter(
            name='vmid', object_types='virtualization.virtualmachine'
        )) > 0
    except Exception as e:
        raise APIError(e)
    if vmid_already_exists:
        raise ValidationError('Custom field "vmid" already exists!')
    custom_fields = [{
        'name': 'vmid',
        'label': 'VMID',
        'description': '[Proxmox] VM/CT ID',
        'ui_editable': 'no',
        'ui_visible': 'always',
        'filter_logic': 'exact',
        'type': 'integer',
        'object_types': ['virtualization.virtualmachine']
    }]
    # Create stuff :)
    try:
        netbox_custom_fields = NB_API.extras.custom_fields.create(custom_fields)
    except Exception as e:
        raise APIError(e)
    # Return JSON serializable list :)
    return [dict(custom_field) for custom_field in netbox_custom_fields]


def vms_and_interfaces():
    proxmox_vms, proxmox_interfaces = Proxmox.fetch_virtual_machines_and_interfaces()
    try:
        netbox_vms = NB_API.virtualization.virtual_machines.create(proxmox_vms)
        netbox_interfaces = NB_API.virtualization.interfaces.create(proxmox_interfaces)
    except Exception as e:
        raise APIError(e)
    # Return JSON serializable lists :)
    return [dict(x) for x in netbox_vms], [dict(x) for x in netbox_interfaces]


def all():
    proxmox_cluster, proxmox_nodes = Proxmox.fetch_cluster_and_nodes()
    _assert_nodes_exist(proxmox_nodes)
    _assert_cluster_does_not_exist(proxmox_cluster)
    # Leave cluster out first so the first error comes quickly, useful for API
    #  errors like invalid token for instance, which might be common.
    # (netbox takes tooooo long to delete <stuff>, even if <stuff> does not exist)
    netbox_cluster = cluster(proxmox_cluster, proxmox_nodes)
    # In case the cluster() call succeeds we are sure that we have write access
    #  for the API, so any error that comes out of the following calls must be
    #  something else (or you messed up things).
    # This is useful because we can be pretty sure that a "rollback" is possible
    # (the delete_everything() call below *should* NOT raise an exception).
    try:
        netbox_tags = tags()
        netbox_custom_fields = custom_fields()
        netbox_vms, netbox_interfaces = vms_and_interfaces()
    except Exception as e:
        # "rollback" in case of any failure
        # (if this function fails you probably messed up in a very interesting way)
        delete_everything()
        raise e
    return {
        'cluster': netbox_cluster,
        'tags': netbox_tags,
        'custom_fields': netbox_custom_fields,
        'vms': netbox_vms,
        'interfaces': netbox_interfaces,
    }
