from netbox_proxmox_sync.api.config import (
    PX_API,
    NETBOX_SITE_ID,
    NETBOX_CLUSTER_DESCRIPTION,
    NETBOX_CLUSTER_TYPE_ID,
)
from netbox_proxmox_sync.api.utils.errors import APIError
from netbox_proxmox_sync.api.utils.models import (
    tag_from_proxmox,
    vm_from_proxmox,
    interface_from_proxmox,
)


# Utils for proper data extraction
def extract_vm_data(cluster_name, node_name, vm_status):
    try:
        vm_config = PX_API.nodes(node_name).qemu(vm_status['vmid']).config.get()
    except Exception:
        raise APIError('Failed to connect to Proxmox Cluster!')
    tags = []
    if vm_config.get('tags') is not None:
        tags = vm_config['tags'].split(';')
    vm_data = vm_from_proxmox(cluster_name, node_name, vm_status, tags)
    interfaces_data = extract_vm_interfaces(vm_config)

    return vm_data, interfaces_data


def extract_vm_interfaces(proxmox_vm_config):
    vm_name = proxmox_vm_config['name']
    interfaces = []
    for interface_name in [key for key in proxmox_vm_config if key.startswith('net')]:
        interface_info = proxmox_vm_config[interface_name]
        new_interface = interface_from_proxmox(vm_name, interface_name, interface_info)
        interfaces.append(new_interface)
    return interfaces


class Proxmox:
    @staticmethod
    def fetch_cluster():
        cluster, _ = Proxmox.fetch_cluster_and_nodes()
        return cluster

    @staticmethod
    def fetch_cluster_and_nodes():
        try:
            proxmox_cluster_info = PX_API.cluster.status.get()
        except Exception:
            raise APIError('Failed to connect to Proxmox Cluster!')
        cluster = {
            'name': proxmox_cluster_info[0]['name'],
            'type': NETBOX_CLUSTER_TYPE_ID,
            'description': NETBOX_CLUSTER_DESCRIPTION,
            'site': NETBOX_SITE_ID,
        }
        return cluster, proxmox_cluster_info[1:]

    @staticmethod
    def fetch_tags():
        try:
            proxmox_tag_info = PX_API.cluster.options.get()
        except Exception:
            raise APIError('Failed to connect to Proxmox Cluster!')
        allowed_tags = proxmox_tag_info['allowed-tags']
        tag_colormap = proxmox_tag_info['tag-style']['color-map'].split(';')
        tags = []
        # Create tags with no color defined
        for tag_name in allowed_tags:
            tags.append(tag_from_proxmox(tag_name))
        # Remap defined tag colors
        for tag_info in tag_colormap:
            tag_name = tag_info.split(':')[0]
            tag_slug = tag_name.lower().replace(' ', '-').replace('.', '_')
            tag_color = tag_info.split(':')[1].lower()
            found = False
            # Find existing tag and update its color
            for tag in tags:
                if tag['name'] == tag_name:
                    found = True
                    tag['color'] = tag_color
                    break
            # ???
            if not found:
                tags.append({
                    'name': tag_name,
                    'slug': tag_slug,
                    'color': tag_color,
                    'object_types': ['virtualization.virtualmachine']
                })
        return tags

    @staticmethod
    def fetch_virtual_machines_and_interfaces():
        try:
            cluster_status = PX_API.cluster.status.get()
        except Exception:
            raise APIError('Failed to connect to Proxmox Cluster!')
        cluster_name = cluster_status[0]['name']
        proxmox_nodes = [node['name'] for node in cluster_status[1:]]
        # List, because node information is stored in the VMs themselves
        # (same for interfaces: vm info is stored in the interface itself)
        virtual_machines = []
        interfaces = []
        for node_name in proxmox_nodes:
            try:
                node_vms = PX_API.nodes(node_name).qemu.get()
            except Exception:
                continue
            for vm_status in node_vms:
                # If the update runs as the VM is being cloned the vm_config will be
                # messed up and our update will fail.
                if vm_status.get('lock') is not None and vm_status['lock'] == 'clone':
                    continue
                new_vm, new_interfaces = extract_vm_data(
                    cluster_name, node_name, vm_status
                )
                virtual_machines.append(new_vm)
                interfaces.extend(new_interfaces)
        return virtual_machines, interfaces
